from .core import add_acast, add_acast_if_supported

try:
    import numpy as np, torch
    add_acast_if_supported(
        "torch", "numpy", "Tensor", "ndarray",
        lambda x: x.detach().cpu().numpy()
    )
    add_acast(np.ndarray, torch.Tensor, torch.from_numpy)
    add_acast(np.ndarray, float, lambda x: float(x.item()) if x.size == 1 else float(x[0]))
except ImportError:
    pass



import pathlib
import importlib
from loguru import logger

from .core import (
    add_acast,
    acaster,
    acast,
    add_acast_if_supported,
    __g_mappings__,
)

__all__ = ["add_acast", "acaster", "acast", "add_acast_if_supported"]

def register_all_backends():
    """Import all backend cast modules (casts_*.py) to register their casts."""
    package_dir = pathlib.Path(__file__).parent

    for file in package_dir.iterdir():
        # Only consider Python files starting with "casts_" and skip __init__.py
        if file.suffix == ".py" and file.name.startswith("casts_"):
            mod_name = f"{__package__}.{file.stem}"
            try:
                importlib.import_module(mod_name)
                logger.debug(f"Successfully imported backend: {mod_name}")
            except ImportError:
                # Expected if used doesn't have the lib installed
                logger.warning(f"Failed to import module: {mod_name}")
                pass
            
register_all_backends()


from .core import add_acast, acast

try:
    import open3d as o3d
    import numpy as np

    def numpy_array_to_vector3d(x):
        array_shape = x.shape
        if len(array_shape) != 2 or array_shape[-1] != 3:
            assert array_shape[-1] == 3, "Input array must have shape (..., 3)"
            x = x.reshape(-1, 3)
        return o3d.utility.Vector3dVector(np.asarray(x))

    add_acast(np.ndarray, o3d.utility.Vector3dVector, numpy_array_to_vector3d)

    def tuple_to_pointcloud(x):
        assert isinstance(x, (tuple, list)) and len(x) == 2, "Input must be (points, colors)"
        pts, cols = x
        pcd = o3d.geometry.PointCloud()
        pcd.points = acast(pts, o3d.utility.Vector3dVector)
        pcd.colors = acast(cols, o3d.utility.Vector3dVector)
        return pcd

    add_acast(tuple, o3d.geometry.PointCloud, tuple_to_pointcloud)
    add_acast(list, o3d.geometry.PointCloud, tuple_to_pointcloud)

except ImportError:
    pass



import pytest
import numpy as np
from autocast import acast, add_acast

try:
    import torch
    HAS_TORCH = True
except ImportError:
    HAS_TORCH = False

@pytest.mark.skipif(not HAS_TORCH, reason="Torch not installed")
def test_torch_to_numpy():
    a = torch.rand(1, 3)
    b = acast(a, np.ndarray)
    assert isinstance(b, np.ndarray)
    assert np.allclose(a.numpy(), b)

def test_dict_recursion():
    a = {"k": torch.rand(1)}
    b = acast(a, float) # Assuming mapping exists or recurses
    # Add your assertions here
    
    
    
# if __name__ == "__main__":
#     print("Testing autocast.py")

#     import json

#     pretty = {str(k): [str(vv) for vv in v.keys()] for k, v in __g_mappings__.items()}
#     print(json.dumps(pretty, indent=4))

#     # import numpy as np
#     # import torch

#     # a = torch.rand(1, 3)
#     # b = acast(a, np.ndarray)
#     # print(a, b)
#     # print(type(a), type(b))

#     # a = {"k": torch.rand(1)}
#     # b = acast(a, float, debug=True)
#     # print(a, b)
#     # print(type(a), type(b))

#     # @acaster(torch.Tensor, int)
#     # def map_torch_to_torch(x=0):
#     #     return int(x + 1)
#     # print(acast(torch.zeros(1), int))

#     # @acaster(torch.Tensor, str)
#     # def map_torch_to_str(x=0):
#     #     import potatoooo
#     #     return "a"
#     # print(acast(torch.zeros(1), str))

#     # path = __find_path__(str, int)
#     # print(path)

#     # path = __find_path__(object, int)
#     # print(path)

#     # path = __find_path__(torch.Tensor, int)
#     # print(path)

#     # a = torch.rand(1, 3)
#     # b = acast(a, np.ndarray, str)
#     # print(a, b)

#     # omegaconf = importlib.import_module("omegaconf")
#     # a = omegaconf.listconfig.ListConfig([1, 2, 3])
#     # b = acast(a, object)
#     # print(a, b, type(a), type(b))

#     a = torch.rand(1)
#     b = acast(a, str, debug=True)
#     print(b)

#     @acaster(torch.Tensor, str)
#     def map_torch_str(x=torch.zeros(1)):
#         print("map_torch_str")
#         return str(x)

#     b = acast(a, str, debug=True)
#     print(b)



import importlib
from collections import defaultdict, deque
from loguru import logger

__g_mappings__ = defaultdict(dict)

def __find_path__(from_type, to_type):
    """Find a path from one type to another using BFS."""
    queue = deque([(from_type, [from_type])])
    visited = set()

    while queue:
        current, path = queue.popleft()

        if current in visited:
            continue

        visited.add(current)

        if current == to_type:
            return path

        for t in __g_mappings__[current]:
            if t not in path:
                queue.append((t, path + [t]))

        # We can try casting from parents of current to type
        for t in current.__bases__:
            if t not in path:
                queue.append((t, path + [t]))

    return None


def add_acast(from_type, to_type, function):
    """Register a function that casts from one type to another."""
    if to_type in __g_mappings__[from_type]:
        # raise ValueError(f"Mapping {from_type} - {to_type} already exists")
        logger.debug(f"Overwriting existing cast: {from_type} -> {to_type}")
    __g_mappings__[from_type][to_type] = function


def acaster(from_type, to_type):
    """Decorator for a function that casts from one type to another.

    Args:
        from_type (_type_): _description_
        to_type (_type_): _description_
    """

    def decorator(func):
        # Run function with default input
        try:
            res = func()
            assert type(res) is to_type, f"Function {func} does not return type {to_type}"
            add_acast(from_type, to_type, func)
            logger.debug(f"Registered acast from {from_type} to {to_type} with {func}")
        except Exception as e:
            logger.error(f"Failed to register acast from {from_type} to {to_type} with {func}: {e}")
        return func
    
    return decorator


def acast(instance, *to_type_list, debug=False):
    from_type = instance.__class__

    if debug:
        logger.debug(f"acast from {from_type} to {to_type_list}")

    for to_type in to_type_list:

        if from_type != to_type:

            path = __find_path__(from_type, to_type)
            if debug:
                logger.debug(f"acast path from {from_type} to {to_type}: {path}")

            if path is not None:

                for i in range(len(path) - 1):
                    s = path[i]
                    t = path[i + 1]

                    if debug:
                        logger.debug(f"acast step {from_type} -> {to_type}: {s} -> {t}")

                    # If mapping exists, use it
                    if s in __g_mappings__ and t in __g_mappings__[s]:
                        instance = __g_mappings__[s][t](instance)
                        # continue

                    # If it's a parent, do nothing
                    # if isinstance(instance, t):
                    #     continue

                from_type = t
                continue

            else:
                # If there is no path, do this for dicts and lists
                # This can be a bit confusing, if we map dict to str, what should we do?
                # a) dict -> obj -> str
                # b) below
                if from_type is dict:
                    return {k: acast(v, *to_type_list, debug=debug) for k, v in instance.items()}

                if from_type is list:
                    return [acast(v, *to_type_list, debug=debug) for v in instance]

            if from_type not in __g_mappings__:
                raise ValueError(f"Mapping from {from_type} not found")

            if to_type not in __g_mappings__[from_type]:
                raise ValueError(f"Mapping to {to_type} not found in {from_type}")

    return instance

def add_acast_if_supported(lib1, lib2, t1, t2, f):
    try:
        lib1 = importlib.import_module(lib1)
        lib2 = importlib.import_module(lib2)
        add_acast(getattr(lib1, t1), getattr(lib2, t2), f)
    except ImportError:
        pass


from .core import add_acast

try:
    import omegaconf
    from omegaconf import OmegaConf
    
    add_acast(omegaconf.listconfig.ListConfig, object, OmegaConf.to_object)
    add_acast(omegaconf.listconfig.ListConfig, list, OmegaConf.to_container)
    
except ImportError:
    pass
